#include "sdlskk_internal.h"
#include <ctype.h>
#include <assert.h>

#define INPUT_BUF_SIZE 2048
#define HENKAN_BUF_SIZE 512
#define OKURIGANA_BUF_SIZE 10
#define PREFIX_BUF_SIZE 10

typedef void (* SDLSKK_Keybind)(SDLSKK_Context* context, char key );

SDLSKK_Context* SDLSKK_Context_new( SDLSKK_Dictionary* dict,
				    SDLSKK_RomKanaRuleTable* table)
{
  SDLSKK_Context* context = SDLSKK_malloc( sizeof(SDLSKK_Context) );
  
  context->dict = dict;
  context->rule_table = table;
  context->mode = SDLSKK_J_KAKUTEI_MODE;
  context->str = SDLSKK_XStr_new4( "", INPUT_BUF_SIZE );
  context->prefix = SDLSKK_XStr_new4( "", PREFIX_BUF_SIZE );
  context->henkan_key = SDLSKK_XStr_new4 ( "", HENKAN_BUF_SIZE );
  context->henkan_okurigana = SDLSKK_XStr_new4( "", OKURIGANA_BUF_SIZE );
  context->dict_item = NULL;
  context->display_str = SDLSKK_XStr_new4( "", INPUT_BUF_SIZE );
  context->minibuffer_str = SDLSKK_XStr_new4( "", INPUT_BUF_SIZE );
  context->cursor_pos = 0;
  context->henkan_buf_pos = 0;
  context->katakana_on = 0;
  
  return context;
}

void SDLSKK_Context_delete( SDLSKK_Context* context )
{
  SDLSKK_XStr_delete( context->str );
  SDLSKK_XStr_delete( context->prefix );
  SDLSKK_XStr_delete( context->henkan_key );
  SDLSKK_XStr_delete( context->henkan_okurigana );
  SDLSKK_XStr_delete( context->display_str );
  SDLSKK_XStr_delete( context->minibuffer_str );
  SDLSKK_DictItem_delete( context->dict_item );
  free( context );
}

void SDLSKK_Context_insert_Char( SDLSKK_Context* context, SDLSKK_Char c )
{
  SDLSKK_XStr_insert_Char( context->str, context->cursor_pos, c);
  ++(context->cursor_pos);
}

static void Context_insert_XStr( SDLSKK_Context* context, SDLSKK_XStr* str )
{
  SDLSKK_XStr_insert_Str( context->str, context->cursor_pos, str->buf);
  context->cursor_pos += str->len;
}

void SDLSKK_Context_insert_Str( SDLSKK_Context* context, SDLSKK_Char *str )
{
  SDLSKK_XStr_insert_Str( context->str, context->cursor_pos, str );
  context->cursor_pos += SDLSKK_Str_len( str );
}

static void SDLSKK_XStr_chop( SDLSKK_XStr* str )
{
  SDLSKK_XStr_delete_Char( str, str->len - 1 );
}

static void Context_henkan_kakutei( SDLSKK_Context* context )
{  
  SDLSKK_Context_insert_Str( context,
			     SDLSKK_DictItem_get_current( context->dict_item ) );
  SDLSKK_Context_insert_Str( context, context->henkan_okurigana->buf );
  SDLSKK_DictItem_commit( context->dict_item );
  SDLSKK_DictItem_delete( context->dict_item );
  SDLSKK_XStr_clear( context->henkan_okurigana );
  context->dict_item = NULL;
  context->mode = SDLSKK_J_KAKUTEI_MODE;
    
}

static SDLSKK_Char* convert_hiragana_katakana( SDLSKK_Char* str, int katanana )
{
  if( katanana )
    return SDLSKK_Str_hiragana_to_katakana( str );
  else
    return SDLSKK_Str_katakana_to_hiragana( str );
}

static SDLSKK_Char* get_kana( SDLSKK_RomKanaElement* kana, int katakana_on )
{
  return katakana_on?kana->katakana:kana->hiragana;
}

static void resolve_romkana( SDLSKK_RomKanaRuleTable* table,
			     SDLSKK_XStr* prefix, SDLSKK_XStr* inserted_str,
			     int* pos, int katakana_on )
{
  SDLSKK_RomKanaElement* kana;
  SDLSKK_XStr* prefix_buf;
  
  if( prefix->len == 0 )
    return;
  
  switch( SDLSKK_RomKanaRuleTable_get( table, prefix->buf, &kana ) ){
    
  case SDLSKK_RomKana_NoCandidate:
    prefix_buf = SDLSKK_XStr_new5( prefix );
    SDLSKK_XStr_chop( prefix_buf );
    
    if( SDLSKK_RomKanaRuleTable_get( table, prefix_buf->buf, &kana ) == 
	SDLSKK_RomKana_UnresolvedCandidate ){
      SDLSKK_Char rest;
      SDLSKK_XStr_insert_Str( inserted_str, *pos,
			      get_kana( kana, katakana_on ) );
      *pos += SDLSKK_Str_len( get_kana( kana, katakana_on ) );
            
     
      rest = prefix->buf[prefix->len-1];
      SDLSKK_XStr_clear( prefix );
      SDLSKK_XStr_insert_Char( prefix, 0, rest );
      
      resolve_romkana( table, prefix, inserted_str, pos, katakana_on );
    }else{
      SDLSKK_XStr_delete_Char( prefix, 0 );
      resolve_romkana( table, prefix, inserted_str, pos, katakana_on );
    }
    SDLSKK_XStr_delete( prefix_buf );
    return;
    
  case SDLSKK_RomKana_SomeCandidates:
    return;
    
  case SDLSKK_RomKana_OneCandidate:
    SDLSKK_XStr_insert_Str( inserted_str, *pos, get_kana( kana, katakana_on ));
    *pos += SDLSKK_Str_len( get_kana( kana, katakana_on ) );

    SDLSKK_XStr_copy_Str( prefix, kana->next_state );
    return;
    
  case SDLSKK_RomKana_UnresolvedCandidate:
    return;
  }
}

static void resolve_finish( SDLSKK_RomKanaRuleTable* rule_table,
			    SDLSKK_XStr* prefix,
			    SDLSKK_XStr* inserted_str,
			    int* pos )
{
  SDLSKK_RomKanaElement* candidate;
  if( SDLSKK_RomKanaRuleTable_get( rule_table, prefix->buf, &candidate ) ==
      SDLSKK_RomKana_UnresolvedCandidate ){
    SDLSKK_XStr_insert_Str( inserted_str, *pos, candidate->hiragana );
    *pos += SDLSKK_Str_len( candidate->hiragana );
    SDLSKK_XStr_clear( prefix );
  }else{
    SDLSKK_XStr_clear( prefix );
  }
}

static void normal_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_insert_Char( context->prefix, context->prefix->len , key );
  resolve_romkana( context->rule_table, context->prefix, context->str,
		   &context->cursor_pos, context->katakana_on  );
}
static void upper_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  SDLSKK_XStr_clear( context->henkan_key );
  SDLSKK_XStr_clear( context->henkan_okurigana );
  context->henkan_buf_pos = 0;
  context->mode = SDLSKK_J_HENKAN_ON_MODE;
  
  SDLSKK_XStr_insert_Char( context->prefix, context->prefix->len,
			   tolower( key ) );      
  resolve_romkana( context->rule_table, context->prefix,
		   context->henkan_key, &(context->henkan_buf_pos),
		   0 );
}  

static void graph_on_latin_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_Context_insert_Char( context, key );
}
static void change_to_latin_mode_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  context->mode = SDLSKK_LATIN_MODE;
  SDLSKK_XStr_clear( context->prefix );
  return;
}
static void change_to_jisx0208_latin_mode_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  context->mode = SDLSKK_JISX0208_LATIN_MODE;
  SDLSKK_XStr_clear( context->prefix );
  return;
}

static void change_to_abbrev_mode_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  context->mode = SDLSKK_ABBREV_MODE;
  SDLSKK_XStr_clear( context->prefix );
  SDLSKK_XStr_clear( context->henkan_key );
  context->henkan_buf_pos = 0;      
  return;
}

static void graph_on_j_henkan_okuri_mode( SDLSKK_Context* context, char key )
{
  int pos;
  
  SDLSKK_XStr_insert_Char( context->prefix, context->prefix->len, key );
  pos = context->henkan_okurigana->len;
  resolve_romkana( context->rule_table, context->prefix,
		   context->henkan_okurigana, &pos, 0 );
  if( context->prefix->len == 0 ){
    SDLSKK_XStr_insert_Char( context->henkan_key, context->henkan_key->len,
			     context->okuri_char );
    context->dict_item = SDLSKK_Dict_get( context->dict,
					  context->henkan_key->buf );
    SDLSKK_XStr_chop( context->henkan_key );
    if( context->dict_item == NULL ){
      context->mode = SDLSKK_J_HENKAN_ON_MODE;
      SDLSKK_XStr_clear( context->henkan_okurigana );
      return;
    }else{
      context->mode = SDLSKK_J_HENKAN_ACTIVE_MODE;
      return;
    }
  }
}
static void upper_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  context->okuri_char = tolower( key );
  SDLSKK_XStr_clear( context->henkan_okurigana );
  context->mode = SDLSKK_J_HENKAN_OKURI_MODE;
  SDLSKK_Context_input_key( context, tolower( key ) );
}
static void normal_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_insert_Char( context->prefix, context->prefix->len, key );
  resolve_romkana( context->rule_table, context->prefix,
		   context->henkan_key, &(context->henkan_buf_pos),
		   0 );
}
static void change_to_latin_mode_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  Context_insert_XStr( context, context->henkan_key );
  SDLSKK_XStr_clear( context->henkan_key );
  context->mode = SDLSKK_LATIN_MODE;
}
static void change_to_jisx0208_latin_mode_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  Context_insert_XStr( context, context->henkan_key );
  SDLSKK_XStr_clear( context->henkan_key );
  context->mode = SDLSKK_JISX0208_LATIN_MODE;
}

static void graph_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  if( islower(key) ){
    normal_on_j_henkan_on_mode( context, key );
  }
  if( isupper(key) ){
    upper_on_j_henkan_on_mode( context, key );
  }
}

static void graph_on_j_henkan_active_mode( SDLSKK_Context* context, char key )
{
  Context_henkan_kakutei( context );
  SDLSKK_Context_input_key( context, key );
}

static void graph_on_abbrev_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_insert_Char( context->henkan_key, context->henkan_buf_pos, key );
  ++(context->henkan_buf_pos);
}

static void graph_on_jisx0208_latin_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_Context_insert_Char( context, SDLSKK_latin_to_jisx0208_Char( key ) );
}

static void toggle_katakana_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  context->katakana_on = !context->katakana_on;
}

static void toggle_kata_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_Char* str;

  resolve_finish( context->rule_table, context->prefix, context->henkan_key,
		  &context->henkan_buf_pos );
  str = convert_hiragana_katakana( context->henkan_key->buf,
				   ! context->katakana_on );
  SDLSKK_Context_insert_Str( context, str );
  free( str );
  SDLSKK_XStr_clear( context->henkan_key );
  context->mode = SDLSKK_J_KAKUTEI_MODE;
}

static void space_on_latin_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_Context_insert_Char( context, ' ' );
}
static void space_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  SDLSKK_Context_insert_Char( context, ' ' );
}


static void space_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  resolve_finish( context->rule_table, context->prefix,
		  context->henkan_key, &context->henkan_buf_pos );
  
  context->dict_item = SDLSKK_Dict_get( context->dict, context->henkan_key->buf );
  if( context->dict_item != NULL )
    context->mode = SDLSKK_J_HENKAN_ACTIVE_MODE;
  
}
static void space_on_j_henkan_active_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_DictItem_next( context->dict_item );
}
static void space_on_abbrev_mode( SDLSKK_Context* context, char key )
{
  context->dict_item = SDLSKK_Dict_get( context->dict, context->henkan_key->buf );
  if( context->dict_item != NULL )
    context->mode = SDLSKK_J_HENKAN_ACTIVE_MODE;
}

static void kakutei_on_latin_mode( SDLSKK_Context* context, char key )
{
  context->mode = SDLSKK_J_KAKUTEI_MODE;
}

static void kakutei_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_Char* kana;
  
  SDLSKK_XStr_clear( context->prefix );
  kana = convert_hiragana_katakana( context->henkan_key->buf,
				    context->katakana_on );
  SDLSKK_Context_insert_Str( context, kana );
  free( kana );
  SDLSKK_XStr_clear( context->henkan_key );
  context->mode = SDLSKK_J_KAKUTEI_MODE;
  return;
}
static void kakutei_on_j_henkan_active_mode( SDLSKK_Context* context, char key )
{
  Context_henkan_kakutei( context );
}

static void kouho_backward_on_j_henkan_active_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_DictItem_prev( context->dict_item );
}
static void forward_on_latin_mode( SDLSKK_Context* context, char key )
{
  if( context->cursor_pos < context->str->len )
    ++(context->cursor_pos);
  return;
}
static void forward_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  if( context->prefix->len > 0 ){
    SDLSKK_XStr_clear( context->prefix );
  }
  if( context->cursor_pos < context->str->len )
    ++(context->cursor_pos);
  return;
}


static void backward_on_latin_mode( SDLSKK_Context* context, char key )
{
  if( context->cursor_pos > 0 )
    --(context->cursor_pos);
  return;
}
static void backward_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  if( context->prefix->len > 0 ){
    SDLSKK_XStr_clear( context->prefix );
  }else if( context->cursor_pos > 0 ){
    --(context->cursor_pos);
  }
  return;
}


static void XStr_backspace( SDLSKK_XStr* str, int* pos )
{
  if( *pos <= 0 )
    return;
  SDLSKK_XStr_delete_Char( str, *pos - 1 );
  --(*pos);
  return;
}
static void backspace_on_latin_mode( SDLSKK_Context* context, char key )
{
  XStr_backspace( context->str, &context->cursor_pos );
  return;
}
static void backspace_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  if( context->prefix->len > 0 ){
    SDLSKK_XStr_clear( context->prefix );
  }else{
    XStr_backspace( context->str, &context->cursor_pos );
  }
  return;
}
static void backspace_on_j_henkan_okuri_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  SDLSKK_XStr_clear( context->henkan_okurigana );
  context->mode = SDLSKK_J_HENKAN_ON_MODE;
}
static void backspace_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  if( context->prefix->len > 0 ){
    SDLSKK_XStr_clear( context->prefix );
  }else if( context->henkan_key->len == 0 ){
    context->mode = SDLSKK_J_KAKUTEI_MODE;
  }else{
    XStr_backspace( context->henkan_key, &context->henkan_buf_pos );
  }
}
static void backspace_on_abbrev_mode( SDLSKK_Context* context, char key )
{
  if( context->henkan_key->len > 0 ){
    XStr_backspace( context->henkan_key, &context->henkan_buf_pos );
  }else{
    context->mode = SDLSKK_J_KAKUTEI_MODE;
  }
}

static void delete_on_latin_mode( SDLSKK_Context* context, char key )
{
  if( context->cursor_pos < context->str->len ){
    context->cursor_pos += 1;
    XStr_backspace( context->str, &context->cursor_pos );
  }
}
static void delete_on_j_kakutei_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  delete_on_latin_mode( context, key );
}

static void cancel_on_j_henkan_on_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  SDLSKK_XStr_clear( context->henkan_key );
  context->mode = SDLSKK_J_KAKUTEI_MODE;
}

static void cancel_on_j_henkan_okuri_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_XStr_clear( context->prefix );
  SDLSKK_XStr_clear( context->henkan_key );
  SDLSKK_XStr_clear( context->henkan_okurigana );
  context->mode = SDLSKK_J_KAKUTEI_MODE;
}

static void cancel_on_j_henkan_active_mode( SDLSKK_Context* context, char key )
{
  SDLSKK_DictItem_delete( context->dict_item );
  context->dict_item = NULL;
  context->mode = SDLSKK_J_HENKAN_ON_MODE;
}

#ifdef HAVE_SDL
#include <SDL.h>
#include "sdlkeybind.tbl"
void SDLSKK_Context_input_event( SDLSKK_Context* context,SDL_Event* event )
{
  if( event->type == SDL_KEYDOWN ){
    SDLSKK_Keybind function;
    function = default_sdl_keybind[event->key.keysym.sym][context->mode];
    if( function != NULL )
      function( context, event->key.keysym.unicode );
  }
}
#endif

#include "keybind.tbl"
void SDLSKK_Context_input_key( SDLSKK_Context* context, char key )
{
  if( default_keybind[key][context->mode] != NULL )
    default_keybind[key][context->mode]( context, key );
}


#ifdef USE_UTF8
#define HENKAN_MARKER (0x25BD)
#define HENKAN_ACTIVE_MARKER (0x25BC)
#else
#define HENKAN_MARKER ((162<<8)+166)
#define HENKAN_ACTIVE_MARKER ((162<<8)+167)
#endif

static void format_henkan_XStr_on_j_henkan_on_mode( SDLSKK_Context* context, SDLSKK_XStr* xstr )
{
  SDLSKK_Char* henkan_key_katakana;
  
  SDLSKK_XStr_insert_Str( xstr, 0, context->prefix->buf );
  if( context->katakana_on ){
    henkan_key_katakana =
      SDLSKK_Str_hiragana_to_katakana( context->henkan_key->buf );
    SDLSKK_XStr_insert_Str( xstr, 0, henkan_key_katakana );
    free( henkan_key_katakana );
  }else{      
    SDLSKK_XStr_insert_Str( xstr, 0, context->henkan_key->buf );
  }
  SDLSKK_XStr_insert_Char( xstr, 0, HENKAN_MARKER );
}

static void format_henkan_XStr_on_j_henkan_okuri_mode( SDLSKK_Context* context, SDLSKK_XStr* xstr )
{
  SDLSKK_Char* henkan_key_katakana;
  SDLSKK_Char* henkan_okurigana_katakana;

  SDLSKK_XStr_insert_Str( xstr, 0, context->prefix->buf );
  if( context->katakana_on ){
    henkan_key_katakana =
      SDLSKK_Str_hiragana_to_katakana( context->henkan_key->buf );
    henkan_okurigana_katakana =
      SDLSKK_Str_hiragana_to_katakana( context->henkan_okurigana->buf );
    SDLSKK_XStr_insert_Str( xstr, 0, henkan_okurigana_katakana );
    SDLSKK_XStr_insert_Char( xstr, 0, '*' );
    SDLSKK_XStr_insert_Str( xstr, 0, henkan_key_katakana );
    free( henkan_key_katakana );
    free( henkan_okurigana_katakana );      
  }else{
    SDLSKK_XStr_insert_Str( xstr, 0, context->henkan_okurigana->buf );
    SDLSKK_XStr_insert_Char( xstr, 0, '*' );
    SDLSKK_XStr_insert_Str( xstr, 0, context->henkan_key->buf );
  }
  SDLSKK_XStr_insert_Char( xstr, 0, HENKAN_MARKER );
}

void SDLSKK_Context_get_henkan_XStr( SDLSKK_Context* context, SDLSKK_XStr* xstr )
{
  SDLSKK_XStr_clear( xstr );
  
  switch( context->mode ){
  case SDLSKK_LATIN_MODE:
    break;
  case SDLSKK_J_KAKUTEI_MODE:
    SDLSKK_XStr_insert_Str( xstr, 0, context->prefix->buf );
    break;
  case SDLSKK_J_HENKAN_ON_MODE:    
    format_henkan_XStr_on_j_henkan_on_mode( context, xstr );
    break;
  case SDLSKK_J_HENKAN_OKURI_MODE:
    format_henkan_XStr_on_j_henkan_okuri_mode( context, xstr );
    break;
  case SDLSKK_J_HENKAN_ACTIVE_MODE:
    SDLSKK_XStr_insert_Str( xstr, 0, context->henkan_okurigana->buf );
    SDLSKK_XStr_insert_Str( xstr, 0, SDLSKK_DictItem_get_current( context->dict_item ) );
    SDLSKK_XStr_insert_Char( xstr, 0, HENKAN_ACTIVE_MARKER );
    break;
  case SDLSKK_JISX0208_LATIN_MODE:
    break;
  case SDLSKK_ABBREV_MODE:
    SDLSKK_XStr_insert_Str( xstr, 0, context->henkan_key->buf );
    SDLSKK_XStr_insert_Char( xstr, 0, HENKAN_MARKER );
    break;
  }
}
SDLSKK_Char* SDLSKK_Context_get_display_Str( SDLSKK_Context* context )
{
  SDLSKK_XStr* converting_str;

  converting_str = SDLSKK_XStr_new4( "", 1024 );
  SDLSKK_XStr_copy_Str( context->display_str, context->str->buf );
  SDLSKK_Context_get_henkan_XStr( context, converting_str );
  SDLSKK_XStr_insert_Str( context->display_str, context->cursor_pos,
			  converting_str->buf );
  SDLSKK_XStr_delete( converting_str );
  return context->display_str->buf;
}
SDLSKK_Char* SDLSKK_Context_get_minibuffer_string( SDLSKK_Context* context );

char* SDLSKK_Context_get_display_str( SDLSKK_Context* context, char* cstr,
				      size_t size )
{
  SDLSKK_Char* str;
  str = SDLSKK_Context_get_display_Str( context );
  return SDLSKK_Str_to_cstr( str, cstr, size );
}
  
char* SDLSKK_Context_get_str( SDLSKK_Context* context, char* cstr,
			      size_t size )
{
  return SDLSKK_Str_to_cstr( context->str->buf, cstr, size );
}
